import textwrap

from airplane.runtime import Run, __execute_internal


def message(
    channel_name: str,
    message: str,  # pylint: disable=redefined-outer-name
    dedent: bool = True,
) -> Run:
    """Runs the builtin message function against a Slack Airplane resource.

    Args:
        channel_name: The slack channel to send a message to.
        message: The message to send to the slack channel.
        dedent: Whether or not to omit leading whitespace from `message`.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the message builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """
    if dedent:
        message = textwrap.dedent(message)
    return __execute_internal(
        "airplane:slack_message",
        {
            "channelName": channel_name,
            "message": message,
        },
        {"slack": "res00000000zteamslack"},
    )



import dataclasses
import datetime
import inspect
from collections import Counter
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import inflection
import typing_extensions
from docstring_parser import parse
from slugify import slugify  # type: ignore

from airplane.config.types import (
    SQL,
    ConfigVar,
    EnvVar,
    File,
    FuncT,
    InputParamTypes,
    LabeledOption,
    LongText,
    ParamConfig,
    ParamType,
    Resource,
    RuntimeType,
    Schedule,
)
from airplane.exceptions import (
    InvalidAnnotationException,
    InvalidTaskConfigurationException,
    UnsupportedDefaultTypeException,
)


@dataclasses.dataclass
class ParamDef:
    """Parameter definition"""

    # ParamDefs have a subset of types that are built in and serializable.
    ParamDefTypes = Union[str, int, float]
    ParamDefOptions = Union[
        List[LabeledOption[str]],
        List[LabeledOption[int]],
        List[LabeledOption[float]],
    ]

    # Represents the original function's argument name.
    # We need this in order to be able to map slug -> arg_name in case they differ.
    arg_name: str
    slug: str
    name: str
    type: ParamType
    description: Optional[str]
    default: Optional[ParamDefTypes]
    required: Optional[bool]
    options: Optional[ParamDefOptions]
    regex: Optional[str]

    DATE_FORMAT = "%Y-%m-%d"
    DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%SZ"
    DATETIME_MILLISECONDS_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

    @staticmethod
    def serialize_param(
        val: InputParamTypes,
    ) -> ParamDefTypes:
        """Transforms a general parameter into a format supported by parameter definition"""
        if isinstance(val, datetime.datetime):
            return val.strftime(ParamDef.DATETIME_FORMAT)
        if isinstance(val, datetime.date):
            return val.strftime(ParamDef.DATE_FORMAT)
        if isinstance(val, ConfigVar):
            return val.name
        return val


@dataclasses.dataclass
class TaskDef:
    """Task definition"""

    func: Callable[..., Any]
    slug: str
    name: str
    runtime: RuntimeType
    entrypoint_func: str
    description: Optional[str]
    require_requests: Optional[bool]
    allow_self_approvals: Optional[bool]
    timeout: Optional[int]
    constraints: Optional[Dict[str, str]]
    resources: Optional[List[Resource]]
    schedules: Optional[List[Schedule]]
    parameters: Optional[List[ParamDef]]
    env_vars: Optional[List[EnvVar]]

    def run(self, params: Dict[str, Any]) -> Any:
        """Execute task function from param dictionary"""
        func_args: Dict[str, Any] = {}
        for param in self.parameters or []:
            # If the user didn't provide a value for the parameter slug
            if param.slug not in params:
                # Fill in None values for optional parameters that aren't provided
                if not param.required:
                    func_args[param.arg_name] = None
                # Otherwise, we fall back to the function default arguments.
            elif param.type == "date":
                func_args[param.arg_name] = datetime.datetime.strptime(
                    params[param.slug], ParamDef.DATE_FORMAT
                ).date()
            elif param.type == "datetime":
                func_args[param.arg_name] = datetime.datetime.strptime(
                    params[param.slug],
                    ParamDef.DATETIME_MILLISECONDS_FORMAT
                    if "." in params[param.slug]
                    else ParamDef.DATETIME_FORMAT,
                )
            elif param.type == "upload":
                func_args[param.arg_name] = File(
                    id=params[param.slug]["id"],
                    url=params[param.slug]["url"],
                )
            elif param.type == "configvar":
                func_args[param.arg_name] = ConfigVar(
                    name=params[param.slug]["name"],
                    value=params[param.slug]["value"],
                )
            else:
                func_args[param.arg_name] = params[param.slug]

        return self.func(**func_args)

    @classmethod
    def build(
        cls,
        func: FuncT,
        runtime: RuntimeType,
        slug: Optional[str],
        name: Optional[str],
        description: Optional[str],
        require_requests: bool,
        allow_self_approvals: bool,
        timeout: int,
        constraints: Optional[Dict[str, str]],
        resources: Optional[List[Resource]],
        schedules: Optional[List[Schedule]],
        env_vars: Optional[List[EnvVar]],
    ) -> "TaskDef":
        """Construct a task definition from a function."""
        task_description = description
        if func.__doc__ is None:
            param_descriptions = {}
        else:
            docstring = parse(func.__doc__)
            param_descriptions = {
                param.arg_name: param.description for param in docstring.params
            }
            task_description = (
                description or docstring.long_description or docstring.short_description
            )
        type_hints = typing_extensions.get_type_hints(func, include_extras=True)
        sig = inspect.signature(func)
        parameters: List[ParamDef] = []
        for param in sig.parameters.values():
            type_hint = type_hints.get(param.name)
            if type_hint is None:
                raise InvalidAnnotationException(
                    prefix="Missing type annotation",
                    func_name=func.__name__,
                    param_name=param.name,
                )
            resolved_type, is_optional, param_config = _resolve_type(
                func.__name__, param.name, type_hint
            )
            if param_config is None:
                param_config = ParamConfig()

            if param.default is inspect.Signature.empty:
                default = None
            else:
                default = ParamDef.serialize_param(param.default)
                if isinstance(default, File):
                    raise UnsupportedDefaultTypeException(
                        "File defaults are not currently supported with inline code configuration."
                    )

            options: Optional[ParamDef.ParamDefOptions]
            if param_config.options is None:
                options = None
            else:
                options = []  # type: ignore
                for option in param_config.options:
                    if isinstance(option, LabeledOption):
                        labeled_option = LabeledOption(
                            label=option.label,
                            value=ParamDef.serialize_param(option.value),
                        )
                    else:
                        value = ParamDef.serialize_param(option)
                        labeled_option = LabeledOption(
                            label=str(value),
                            value=value,
                        )
                    options.append(labeled_option)  # type: ignore

            default_slug = make_slug(param.name)
            parameters.append(
                ParamDef(
                    arg_name=param.name,
                    # Parameter slug is the parameter's name in snakecase
                    slug=param_config.slug or default_slug,
                    name=param_config.name or inflection.humanize(default_slug),
                    type=_to_airplane_type(func.__name__, param.name, resolved_type),
                    description=param_config.description
                    or param_descriptions.get(param.name),
                    required=not is_optional,
                    default=default,
                    options=options,
                    regex=param_config.regex,
                )
            )

        def _check_duplicates(counter: Counter, duplicate_type: str) -> None:
            duplicates = [slug for slug, count in counter.items() if count > 1]
            if duplicates:
                raise InvalidTaskConfigurationException(
                    f"Function {func.__name__} has duplicate {duplicate_type} {duplicates}"
                )

        _check_duplicates(Counter([p.slug for p in parameters]), "parameter slugs")
        _check_duplicates(Counter([s.slug for s in schedules or []]), "schedule slugs")
        _check_duplicates(Counter([e.name for e in env_vars or []]), "env var names")

        # Convert schedule param values to the correct default types
        for schedule in schedules or []:
            if schedule.param_values is not None:
                for param_name, param_value in schedule.param_values.items():
                    if param_value is not None:
                        schedule.param_values[param_name] = ParamDef.serialize_param(
                            param_value
                        )

        # pylint: disable=protected-access
        default_slug = make_slug(func.__name__)
        return cls(
            func=func,
            runtime=runtime,
            slug=slug or default_slug,
            name=name or inflection.humanize(default_slug),
            description=task_description,
            require_requests=require_requests,
            allow_self_approvals=allow_self_approvals,
            timeout=timeout,
            constraints=constraints,
            schedules=schedules,
            resources=resources,
            parameters=parameters,
            env_vars=env_vars,
            entrypoint_func=func.__name__,
        )


def _resolve_type(
    func_name: str, param_name: str, type_hint: Any
) -> Tuple[Any, bool, Optional[ParamConfig]]:
    """Parses a parameter's type hint to extract its underlying type,
    whether it's optional, and a ParamConfig if provided."""
    origin_type = typing_extensions.get_origin(type_hint)
    if origin_type == Union:
        type_args = typing_extensions.get_args(type_hint)
        if len(type_args) != 2 or type_args[1] is not type(None):
            raise InvalidAnnotationException(
                prefix=f"Unsupported Union type `{type_hint}`",
                func_name=func_name,
                param_name=param_name,
            )
        return type_args[0], True, None
    if origin_type == typing_extensions.Annotated:
        type_args = typing_extensions.get_args(type_hint)
        param_configs = [t for t in type_args if isinstance(t, ParamConfig)]
        # Don't support multiple parameter configs within the same annotation.
        if len(param_configs) > 1:
            raise InvalidAnnotationException(
                prefix=f"Found multiple ParamConfig annotations `{type_hint}`",
                func_name=func_name,
                param_name=param_name,
            )
        param_config = param_configs[0] if param_configs else None
        underlying_type, is_optional, _ = _resolve_type(
            func_name, param_name, type_args[0]
        )
        return underlying_type, is_optional, param_config
    return type_hint, False, None


def _to_airplane_type(func_name: str, param_name: str, type_hint: Any) -> ParamType:
    if type_hint == str:
        return "shorttext"
    if type_hint == LongText:
        return "longtext"
    if type_hint == SQL:
        return "sql"
    if type_hint == bool:
        return "boolean"
    if type_hint == File:
        return "upload"
    if type_hint == int:
        return "integer"
    if type_hint == float:
        return "float"
    if type_hint == datetime.date:
        return "date"
    if type_hint == datetime.datetime:
        return "datetime"
    if type_hint == ConfigVar:
        return "configvar"
    raise InvalidAnnotationException(
        prefix=f"Invalid type annotation `{type_hint}`",
        func_name=func_name,
        param_name=param_name,
    )


def make_slug(string: str) -> str:
    """Turns a string into a slug"""
    for target, replacement in [
        ("‒", "_"),  # figure dash
        ("–", "_"),  # en dash
        ("—", "_"),  # em dash
        ("―", "_"),  # horizontal bar
        ("&", "_and_"),
        ("@", "_at_"),
        ("%", "_percent_"),
    ]:
        string = string.replace(target, replacement)
    return slugify(inflection.underscore(string)).replace("-", "_")[:50]



import textwrap
from dataclasses import asdict, dataclass, is_dataclass
from typing import List, Union

from airplane.builtins import __convert_resource_alias_to_id
from airplane.runtime import Run, __execute_internal


@dataclass
class Contact:
    """Representation of an email contact (sender or recipient).

    Attributes:
        email: The email of the contact.
        name: The name of the contact.
    """

    email: str
    name: str


def message(
    email_resource: str,
    sender: Contact,
    recipients: Union[List[Contact], List[str]],
    subject: str = "",
    message: str = "",  # pylint: disable=redefined-outer-name
    dedent: bool = True,
) -> Run:
    """Runs the builtin message function against an email Airplane resource.

    Args:
        email_resource: The alias of the email resource to send the email with.
        sender: The email's sender information.
        recipients: List of the email's recipients.
        subject: The subject of the email.
        message: The message body of the email.
        dedent: Whether or not to omit leading whitespace from `message`.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the message builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """
    if dedent:
        message = textwrap.dedent(message)
    return __execute_internal(
        "airplane:email_message",
        {
            "sender": asdict(sender),
            "recipients": [
                asdict(recipient) if is_dataclass(recipient) else recipient
                for recipient in recipients
            ],
            "subject": subject,
            "message": message,
        },
        {"email": __convert_resource_alias_to_id(email_resource)},
    )



import textwrap
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Set, Union

import deprecation

from airplane._version import __version__
from airplane.api.client import api_client_from_env


def text(content: str, dedent: bool = True) -> None:
    """Appends a display to the run that renders the provided markdown text.

    See the [CommonMark docs](https://commonmark.org/help/) for an introduction
    to markdown formatting.

    Args:
        content: Text to render as markdown
        dedent: Whether or not to omit leading whitespace from content.

    Raises:
        HTTPError: If the display could not be created.
    """
    if dedent:
        content = textwrap.dedent(content)
    api_client_from_env().create_text_display(content)


@deprecation.deprecated(
    deprecated_in="0.3.14",
    current_version=__version__,
    details="Use text(content, dedent) instead.",
)
def markdown(content: str, dedent: bool = True) -> None:
    """Appends a display to the run that renders the provided markdown text.

    See the [CommonMark docs](https://commonmark.org/help/) for an introduction
    to markdown formatting.

    Args:
        content: Text to render as markdown
        dedent: Whether or not to omit leading whitespace from content.

    Raises:
        HTTPError: If the display could not be created.
    """
    text(content, dedent)


def json(
    payload: Union[Dict[str, Any], List[Any], int, str, float, bool, None],
) -> None:
    """Appends a display to the run that renders a JSON payload.

    Args:
        payload: JSON payload to render

    Raises:
        HTTPError: If the display could not be created.
    """
    api_client_from_env().create_json_display(payload)


@dataclass(frozen=True)
class TableColumn:
    """Column for a table display.

    Attributes:
        slug: Identifier used to reference this column.
        name: Column display name. Defaults to slug.
    """

    slug: str
    name: Optional[str] = None


def table(
    rows: List[Dict[str, Any]], columns: Optional[List[Union[str, TableColumn]]] = None
) -> None:
    """Appends a display to the run that renders a table.

    Each row should be an object mapping header slugs to values. Columns that are not
    specified will default to `null`. The selection, ordering, and naming of columns
    can be customized via `opts.columns`.

    Args:
        rows:
            The list of rows to render in the table.
        columns:
            The list of columns to include in the table.

            The order of columns in this list determines the order of columns when rendering the
            table. Each column can optionally specify a human-readable name that will be used when
            rendering the table. The name defaults to the slug.

            Columns found in `rows` that not included in `columns` will not be rendered.

            If not specified, columns are inferred automatically from the provided rows:
            - The set of columns is the union of all keys across all rows.
            - The column order is inferred from the key order of the first row. All other columns
              not present in the first row are ordered after.
            - Columns are named by their slug.

    Raises:
        HTTPError: If the display could not be created.
        ValueError: If columns have empty or duplicate slugs.
    """
    table_columns: List[TableColumn]
    if columns is None:
        # Use a list to maintain dict key order.
        slugs = []
        all_slugs = set()
        for row in rows:
            for key in row.keys():
                if key not in all_slugs:
                    slugs.append(key)
                    all_slugs.add(key)
        table_columns = [TableColumn(slug=s) for s in slugs]
    else:
        table_columns = [
            TableColumn(slug=column) if isinstance(column, str) else column
            for column in columns
        ]
        if any(c.slug == "" for c in table_columns):
            raise ValueError("Column slugs cannot be empty")
        slug_set = {c.slug for c in table_columns}
        if len(columns) != len(slug_set):
            raise ValueError("Column slugs must be unique")
        rows = _filter_row_keys(rows, slug_set)
    api_client_from_env().create_table_display(
        columns=[{"slug": c.slug, "name": c.name} for c in table_columns], rows=rows
    )


def _filter_row_keys(
    rows: List[Dict[str, Any]], columns: Set[str]
) -> List[Dict[str, Any]]:
    filtered_rows = []
    for row in rows:
        filtered_rows.append(dict((c, row[c]) for c in columns if c in row))
    return filtered_rows



import textwrap
from enum import Enum
from typing import Any, Dict, Optional

from airplane.builtins import __convert_resource_alias_to_id
from airplane.runtime import Run, __execute_internal


class TransactionMode(Enum):
    """Valid transaction modes for SQL Airplane resources."""

    AUTO = "auto"
    READ_ONLY = "readOnly"
    READ_WRITE = "readWrite"
    NONE = "none"


def query(
    sql_resource: str,
    query: str,  # pylint: disable=redefined-outer-name
    query_args: Optional[Dict[str, Any]] = None,
    transaction_mode: TransactionMode = TransactionMode.AUTO,
    dedent: bool = True,
) -> Run:
    """Runs the builtin query function against a SQL Airplane resource.

    Args:
        sql_resource: The alias of the SQL resource to execute the query against.
        query: The query to run on the SQL resource.
        query_args: Optional map of query arg names to values to insert into the query.
        transaction_mode: Optional transaction mode with which to run the query.
        dedent: Whether or not to omit leading whitespace from `query`.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the query builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """
    if dedent:
        query = textwrap.dedent(query)
    return __execute_internal(
        "airplane:sql_query",
        {
            "query": query,
            "queryArgs": query_args,
            "transactionMode": transaction_mode.value,
        },
        {"db": __convert_resource_alias_to_id(sql_resource)},
    )



import dataclasses
import datetime
import typing
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    NewType,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

from typing_extensions import Literal

from airplane.exceptions import InvalidTaskConfigurationException

ParamType = Literal[
    "shorttext",
    "longtext",
    "sql",
    "boolean",
    "upload",
    "integer",
    "float",
    "date",
    "datetime",
    "configvar",
]


RuntimeType = Literal["", "workflow"]


if typing.TYPE_CHECKING:
    LongText = str
    SQL = str
else:
    # This is needed to differentiate LongText / SQL from str when building
    # the definition otherwise the label `param: LongText` would be indistinguishable
    # from str. We only want to do this at runtime in order to allow users to still
    # assign strings as default values without have to wrap their types,
    # e.g. param: LongText = "foo"
    LongText = NewType("LongText", str)
    SQL = NewType("SQL", str)


@dataclasses.dataclass
class File:
    """Airplane file parameter.

    File uploads are serialized as an object when passed to tasks.
    https://docs.airplane.dev/platform/parameters#file

    NOTE: Inline task definitions are currently unable to set default File parameters.

    Attributes:
        id:
            File upload ID.
        url:
            Signed URL that can be used to access the uploaded file.
            An HTTP GET to this URL will respond with the uploaded file encoded as a
            byte stream in the response's body.
    """

    id: str
    url: str


@dataclasses.dataclass
class ConfigVar:
    """Airplane config variable parameter.

    Configs variables allow users to set team-wide values / secrets
    and use them in tasks. https://docs.airplane.dev/platform/parameters#config-variable
    """

    name: str
    value: str


@dataclasses.dataclass
class Resource:
    """Airplane resource attachment.

    Resources in Airplane allow users to configure connections to external systems
    like databases and APIs and use them in tasks and runbooks.
    https://docs.airplane.dev/resources/overview

    Attributes:
        slug:
            Resource identifier.
        alias:
            Alias to reference the resource. Defaults to resource slug.
    """

    slug: str
    alias: Optional[str] = None


InputParamTypes = Union[
    str,
    LongText,
    SQL,
    bool,
    # File is not included since we don't allow users to provide default File values
    # via inline configuration yet.
    # File,
    int,
    float,
    datetime.date,
    datetime.datetime,
    ConfigVar,
]
DefaultParamT = TypeVar(
    "DefaultParamT",
    bound=InputParamTypes,
)


@dataclasses.dataclass
class LabeledOption(Generic[DefaultParamT]):
    """Parmeter select option with a label."""

    label: str
    value: DefaultParamT


OptionsT = Sequence[Union[DefaultParamT, LabeledOption[DefaultParamT]]]
AllOptions = Union[
    OptionsT[str],
    OptionsT[int],
    OptionsT[float],
    OptionsT[datetime.date],
    OptionsT[datetime.datetime],
    OptionsT[ConfigVar],
]


@dataclasses.dataclass
class Schedule:
    """Airplane schedule definition.

    Schedules allow users to automatically run tasks on a recurring schedule.
    https://docs.airplane.dev/schedules/schedules

    Attributes:
        slug:
            Human-friendly identifier used for the schedule. Schedule slugs must be unique
            within an individual task / workflow.
        cron:
            Schedule cron expression, e.g.  "0 * * * *"
        name:
            Schedule name. Defaults to the slug.
        description:
            Schedule description displayed on the Airplane app.
        param_values:
            Dictionary of parameter name to parameter value used for the scheduled execution.
            All required parameters must be defined.
    """

    slug: str
    cron: str
    name: Optional[str] = None
    description: Optional[str] = None
    param_values: Optional[Dict[str, Optional[InputParamTypes]]] = None


@dataclasses.dataclass
class EnvVar:
    """Airplane environment variable.

    Environment variables allow users to configure constant values or reference
    config variables. They can be accessed via `os.getenv("MY_ENV_VAR_NAME")`.

    Attributes:
        name:
            Environment variable name.
        value:
            Set a constant value for the environment variable.
        config_var_name:
            Name of the config variable to use. Configs variables allow users
            to set team-wide values / secrets.
    """

    name: str
    value: Optional[str] = None
    config_var_name: Optional[str] = None

    def __post_init__(self) -> None:
        if "=" in self.name:
            raise InvalidTaskConfigurationException(
                f'Environment variable name {self.name} cannot contain "="'
            )
        if (self.config_var_name and self.value) or (
            not self.config_var_name and not self.value
        ):
            raise InvalidTaskConfigurationException(
                "Exactly one of `config_var_name` or `value` should be set"
            )


@dataclasses.dataclass
class ParamConfig:
    """Task parameter configuration.

    Attributes:
        slug:
            Human-friendly identifier used to reference this parameter. Parameter slugs
            must be unique within an individual task / workflow. Defaults to the function
            argument's name.
        name:
            Parameter name displayed on the Airplane app. Defaults to the function
            argument's name in sentence case.
        description:
            Parameter description displayed on the Airplane app. If not provided, the description
            will be pulled from the docstring of the decorated function.
        options:
            Option constraint for the parameter. Select options allow users to specify exactly
            which values are allowed for a given parameter. Options may specify a label which
            is shown to the user on the Airplane app instead of the value.
        regex:
            Regex contraint for the parameter, only valid for string arguments.
    """

    slug: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = None
    options: Optional[AllOptions] = None
    regex: Optional[str] = None


FuncT = TypeVar("FuncT", bound=Callable[..., Any])



"""airplane - An SDK for writing Airplane tasks in Python"""

from airplane import display
from airplane._version import __version__
from airplane.api.client import APIClient
from airplane.api.entities import Run, RunStatus
from airplane.builtins import email, graphql, mongodb, rest, slack, sql
from airplane.config.config import task
from airplane.config.types import (
    SQL,
    ConfigVar,
    EnvVar,
    File,
    LabeledOption,
    LongText,
    ParamConfig,
    Resource,
    Schedule,
)
from airplane.exceptions import InvalidEnvironmentException, RunPendingException
from airplane.output import append_output, set_output, write_named_output, write_output
from airplane.runtime import execute

# Deprecated
from airplane.runtime.standard import run



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional


class RunStatus(Enum):
    """Valid statuses during a run's lifecycle."""

    NOT_STARTED = "NotStarted"
    QUEUED = "Queued"
    ACTIVE = "Active"
    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    CANCELLED = "Cancelled"

    def is_terminal(self) -> bool:
        """Returns whether the status is terminal.

        Returns:
            Whether the status is terminal.
        """

        return self in [self.SUCCEEDED, self.FAILED, self.CANCELLED]


@dataclass
class Run:
    """Representation of an Airplane run.

    Attributes:
        id: The id of the run.
        task_id: The task id associated with the run (None for builtin tasks).
        param_values: The param values the run was provided.
        status: The current status of the run.
        output: The outputs (if any) of the run.
    """

    id: str
    task_id: Optional[str]
    param_values: Dict[str, Any]
    status: RunStatus
    output: Any



from typing import Any, Dict, Optional

import backoff
import deprecation
import requests

from airplane._version import __version__
from airplane.api.client import api_client_from_env
from airplane.api.entities import Run, RunStatus
from airplane.config.types import InputParamTypes
from airplane.exceptions import RunPendingException, RunTerminationException


def execute(
    slug: str,
    param_values: Optional[Dict[str, InputParamTypes]] = None,
    resources: Optional[Dict[str, Any]] = None,
) -> Run:
    """Standard executes an Airplane task, waits for execution, and returns run metadata.

    Args:
        slug: The slug of the task to run.
        param_values: Optional map of parameter slugs to values.
        resources: Optional map of resource aliases to ids.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the task cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    client = api_client_from_env()
    run_id = client.execute_task(slug, param_values, resources)
    run_info = __wait_for_run_completion(run_id)
    outputs = client.get_run_output(run_id)
    # pylint: disable=redefined-outer-name
    run = Run(
        id=run_info["id"],
        task_id=run_info.get("taskID", None),
        param_values=run_info["paramValues"],
        status=RunStatus(run_info["status"]),
        output=outputs,
    )

    if run.status in (RunStatus.FAILED, RunStatus.CANCELLED):
        raise RunTerminationException(run)

    return run


@deprecation.deprecated(
    deprecated_in="0.3.2",
    current_version=__version__,
    details="Use execute(slug, param_values) instead.",
)
def run(
    task_id: str,
    parameters: Optional[Dict[str, Any]] = None,
    env: Optional[Dict[str, Any]] = None,
    constraints: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """Creates an Airplane run, waits for execution, and returns its output and status.

    Args:
        task_id: The id of the task to run.
        parameters: Optional map of parameter slugs to values.
        env: Optional map of environment variables.
        constraints: Optional map of run constraints.

    Returns:
        The status and outputs of the run.

    Raises:
        HTTPError: If the run cannot be created or executed properly.
    """
    client = api_client_from_env()
    run_id = client.create_run(task_id, parameters, env, constraints)
    run_info = __wait_for_run_completion(run_id)
    outputs = client.get_run_output(run_id)
    return {"status": run_info["status"], "outputs": outputs}


@backoff.on_exception(
    lambda: backoff.expo(factor=0.1, max_value=5),
    (
        requests.exceptions.ConnectionError,
        requests.exceptions.Timeout,
        RunPendingException,
    ),
)
def __wait_for_run_completion(run_id: str) -> Dict[str, Any]:
    client = api_client_from_env()
    run_info = client.get_run(run_id)
    if run_info["status"] in ("NotStarted", "Queued", "Active"):
        raise RunPendingException()
    return run_info



from typing import Any, Dict, List, Optional

from airplane.builtins import __convert_resource_alias_to_id
from airplane.runtime import Run, __execute_internal


def find(
    mongodb_resource: str,
    collection: str,
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
    skip: Optional[int] = None,
    limit: Optional[int] = None,
) -> Run:
    """Runs the find function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search in.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.
        skip: Number of documents to skip.
        limit: The maximum number of documents to return.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the find builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_find",
        {
            "collection": collection,
            "filter": filter,
            "projection": projection,
            "sort": sort,
            "skip": skip,
            "limit": limit,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def find_one(
    mongodb_resource: str,
    collection: str,
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
) -> Run:
    """Runs the findOne function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the findOne builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_findOne",
        {
            "collection": collection,
            "filter": filter,
            "projection": projection,
            "sort": sort,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def find_one_and_delete(
    mongodb_resource: str,
    collection: str,
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
) -> Run:
    """Runs the findOneAndDelete function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search in.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the findOneAndDelete builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_findOneAndDelete",
        {
            "collection": collection,
            "filter": filter,
            "projection": projection,
            "sort": sort,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def find_one_and_update(
    mongodb_resource: str,
    collection: str,
    update: Dict[str, Any],
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
) -> Run:
    """Runs the findOneAndUpdate function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search in.
        update: The update document.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the findOneAndUpdate builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_findOneAndUpdate",
        {
            "collection": collection,
            "update": update,
            "filter": filter,
            "projection": projection,
            "sort": sort,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def find_one_and_replace(
    mongodb_resource: str,
    collection: str,
    replacement: Dict[str, Any],
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    projection: Optional[Dict[str, Any]] = None,
    sort: Optional[Dict[str, Any]] = None,
    upsert: Optional[bool] = None,
) -> Run:
    """Runs the findOneAndReplace function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search in.
        replacement: The replacement document.
        filter: The query predicate.
        projection: The projection specification that determines which fields to return.
        sort: The sort specification for the ordering of the results.
        upsert: Replaces the document or inserts the replacement if no document is found.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the findOneAndReplace builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_findOneAndReplace",
        {
            "collection": collection,
            "replacement": replacement,
            "filter": filter,
            "projection": projection,
            "sort": sort,
            "upsert": upsert,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def insert_one(
    mongodb_resource: str,
    collection: str,
    document: Dict[str, Any],
) -> Run:
    """Runs the insertOne function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to insert in.
        document: The document to insert.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the insertOne builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_insertOne",
        {
            "collection": collection,
            "document": document,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def insert_many(
    mongodb_resource: str,
    collection: str,
    documents: List[Dict[str, Any]],
) -> Run:
    """Runs the insertMany function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to insert in.
        documents: The documents to insert.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the insertMany builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_insertMany",
        {
            "collection": collection,
            "documents": documents,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def update_one(
    mongodb_resource: str,
    collection: str,
    update: Dict[str, Any],
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    upsert: Optional[bool] = None,
) -> Run:
    """Runs the updateOne function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to update in.
        update: The update document.
        filter: The query predicate.
        upsert: Updates the document or inserts the update if no document is found.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the updateOne builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_updateOne",
        {
            "collection": collection,
            "update": update,
            "filter": filter,
            "upsert": upsert,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def update_many(
    mongodb_resource: str,
    collection: str,
    update: Dict[str, Any],
    filter: Optional[Dict[str, Any]] = None,  # pylint: disable=redefined-builtin
    upsert: Optional[bool] = None,
) -> Run:
    """Runs the updateMany function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to update in.
        update: The update document.
        filter: The query predicate.
        upsert: Updates the documents or inserts the update if no document is found.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the updateMany builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_updateMany",
        {
            "collection": collection,
            "update": update,
            "filter": filter,
            "upsert": upsert,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def delete_one(
    mongodb_resource: str,
    collection: str,
    filter: Dict[str, Any],  # pylint: disable=redefined-builtin
) -> Run:
    """Runs the deleteOne function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to delete in.
        filter: The query predicate.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the deleteOne builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_deleteOne",
        {
            "collection": collection,
            "filter": filter,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def delete_many(
    mongodb_resource: str,
    collection: str,
    filter: Dict[str, Any],  # pylint: disable=redefined-builtin
) -> Run:
    """Runs the deleteMany function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to delete in.
        filter: The query predicate.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the deleteMany builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_deleteMany",
        {
            "collection": collection,
            "filter": filter,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def aggregate(
    mongodb_resource: str,
    collection: str,
    pipeline: List[Dict[str, Any]],
) -> Run:
    """Runs the aggregate function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to aggregate in.
        pipeline: The sequence of data aggregation operations.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the aggregate builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_aggregate",
        {
            "collection": collection,
            "pipeline": pipeline,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def count_documents(
    mongodb_resource: str,
    collection: str,
    filter: Dict[str, Any],  # pylint: disable=redefined-builtin
) -> Run:
    """Runs the countDocuments function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to count in.
        filter: The query predicate.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the countDocuments builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_countDocuments",
        {
            "collection": collection,
            "filter": filter,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )


def distinct(
    mongodb_resource: str,
    collection: str,
    field: str,
    filter: Dict[str, Any],  # pylint: disable=redefined-builtin
) -> Run:
    """Runs the distinct function against a MongoDB Airplane resource.

    Args:
        mongodb_resource: The alias of the MongoDB resource to use.
        collection: The collection to search.
        field: The field for which to return distinct values.
        filter: The query predicate.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the distinct builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:mongodb_distinct",
        {
            "collection": collection,
            "field": field,
            "filter": filter,
        },
        {"db": __convert_resource_alias_to_id(mongodb_resource)},
    )



import os
from enum import Enum
from typing import Any, Dict, Optional

from airplane.api.entities import Run
from airplane.runtime.standard import execute as standard_execute

__AIRPLANE_RUNTIME_ENV_VAR = "AIRPLANE_RUNTIME"


class RuntimeKind(Enum):
    """Valid runtime kinds for Airplane runs."""

    DEV = "dev"
    STANDARD = ""
    WORKFLOW = "workflow"


def execute(slug: str, param_values: Optional[Dict[str, Any]] = None) -> Run:
    """Executes an Airplane task, waits for execution, and returns run metadata.

    Args:
        slug: The slug of the task to run.
        param_values: Optional map of parameter slugs to values.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the task cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
        NotImplementedError: For workflow runs.
    """
    return __execute_internal(slug, param_values)


def __execute_internal(
    slug: str,
    param_values: Optional[Dict[str, Any]] = None,
    resources: Optional[Dict[str, Any]] = None,
) -> Run:
    runtime_kind = os.environ.get(
        __AIRPLANE_RUNTIME_ENV_VAR, RuntimeKind.STANDARD.value
    )

    if runtime_kind == RuntimeKind.WORKFLOW.value:
        raise NotImplementedError("Workflow run not supported yet by python sdk")

    return standard_execute(slug, param_values, resources)



import functools
from typing import Any, Callable, Dict, List, Optional

from airplane.api.entities import Run
from airplane.config.definitions import TaskDef
from airplane.config.types import EnvVar, FuncT, Resource, Schedule
from airplane.runtime import execute


def task(
    slug: Optional[str] = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    require_requests: bool = False,
    allow_self_approvals: bool = True,
    timeout: int = 3600,
    constraints: Optional[Dict[str, str]] = None,
    resources: Optional[List[Resource]] = None,
    schedules: Optional[List[Schedule]] = None,
    env_vars: Optional[List[EnvVar]] = None,
) -> Callable[[FuncT], Callable[..., Run]]:
    """Decorator used to define an Airplane task.

    This decorator inspects the decorated function to create an Airplane task. The task's parameters
    are deduced from the function's arguments and type hints.

    Additional parameter metadata can be provided by attaching an airplane.ParamConfig to a
    typing.Annotated typehint (for versions of Python prior to 3.9, typing_extensions.Annotated
    can be used).

    The following map enumerates the supported Python types and equivalent Airplane
    parameter types (Airplane parameter details https://docs.airplane.dev/platform/parameters)::

        str: short text
        airplane.LongText: long text
        airplane.SQL: SQL
        bool: boolean
        int: integer
        float: number
        airplane.File: file
        datetime.date: date
        datetime.datetime: datetime
        airplane.ConfigVar: config variable

    Example:
        Sample task::

            @airplane.task()
            def add_two_numbers(first_number: int, second_number: int) -> int:
                return first_number + second_number

        Sample task with parameter annotation::

            @airplane.task()
            def capitalize_string(
                input: Annotated[str, airplane.ParamConfig(name="User string")]
            ) -> str:
                return input.capitalize()
    Args:
        slug:
            Human-friendly identifier used to reference this task. Must be unique
            across tasks and workflows. Defaults to function name.
        name:
            Task name displayed on the Airplane app. Defaults to funcion name in sentence case.
        description:
            Task description displayed on the Airplane app. If not provided, the description
            will be pulled from the docstring of the decorated function.
        require_requests:
            Whether or not this task requires a request to execute.
        allow_self_approvals:
            Whether or not this task allows self approvals.
        timeout:
            How long a task can run (in seconds) for before it is automatically cancelled.
        constraints:
            Constraints for which agents are allowed to execute this task's runs, only
            applicable for users with self hosted agents.
        resources:
            Resources to attach to this task. Resources can be accessed through environment
            variables or built-ins. Resources accessed by this task must be explicitly attached
            in the task's definition.
        schedules:
            Schedules to attach to this task. Schedules allow users to automatically run
            task on a recurring schedule.
        env_vars:
            Enviornment variables to attach to this task. Environment variables allow users
            to configure constant values or reference config variables.
    """

    def decorator(func: FuncT) -> Callable[..., Run]:
        """Assigns an __airplane attribute to a function to mark it as an Airplane object"""

        config = TaskDef.build(
            func=func,
            runtime="",
            slug=slug,
            name=name,
            description=description,
            require_requests=require_requests,
            allow_self_approvals=allow_self_approvals,
            timeout=timeout,
            constraints=constraints,
            resources=resources,
            schedules=schedules,
            env_vars=env_vars,
        )

        @functools.wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> Run:
            kwargs.update(zip(func.__code__.co_varnames, args))
            return execute(config.slug, kwargs)

        # pylint: disable=protected-access
        wrapped.__airplane = config  # type: ignore
        return wrapped

    return decorator



from typing import Any, Dict, Optional

from airplane.builtins import __convert_resource_alias_to_id
from airplane.runtime import Run, __execute_internal


def request(
    graphql_resource: str,
    operation: str,
    variables: Optional[Dict[str, Any]] = None,
    headers: Optional[Dict[str, Any]] = None,
    url_params: Optional[Dict[str, Any]] = None,
    retry_failures: bool = False,
) -> Run:
    """Runs the builtin request function against a GraphQL Airplane resource.

    Args:
        graphql_resource: The alias of the GraphQL resource to use.
        operation: The GraphQL operation to execute.
        variables: Optional GraphQL variables to include in the request.
        headers: Optional headers to include in the request.
        url_params: Optional url params to include in the request.
        retry_failures: True to retry 500, 502, 503, and 504 error codes.

    Returns:
        The id, task id, param values, status and outputs of the executed run.

    Raises:
        HTTPError: If the request builtin cannot be executed properly.
        RunTerminationException: If the run fails or is cancelled.
    """

    return __execute_internal(
        "airplane:graphql_request",
        {
            "operation": operation,
            "variables": variables,
            "headers": headers,
            "urlParams": url_params,
            "retryFailures": retry_failures,
        },
        {"api": __convert_resource_alias_to_id(graphql_resource)},
    )
